unit unit_2D_polynomial_fit;

{$mode Delphi}

interface

uses
  Classes, SysUtils,math;

implementation


type
  TVector = array of Double;
  TMatrix = array of TVector;

//var
//  ErrorMatrix :tmatrix;
//  imageWidth, imageHeight : integer;  // Height of the image

function CreateMatrix(Rows, Cols: Integer): TMatrix;
var
  i: Integer;
begin
  SetLength(Result, Rows);
  for i := 0 to Rows - 1 do
    SetLength(Result[i], Cols);
end;

function TransposeMatrix(const A: TMatrix): TMatrix;
var
  i, j: Integer;
begin
  Result := CreateMatrix(Length(A[0]), Length(A));
  for i := 0 to High(A) do
    for j := 0 to High(A[i]) do
      Result[j][i] := A[i][j];
end;

function MultiplyMatrices(const A, B: TMatrix): TMatrix;
var
  i, j, k: Integer;
begin
  if Length(A[0]) <> Length(B) then
    raise Exception.Create('Matrix dimensions must agree.');

  Result := CreateMatrix(Length(A), Length(B[0]));
  for i := 0 to High(Result) do
    for j := 0 to High(Result[i]) do
      for k := 0 to High(A[0]) do
        Result[i][j] := Result[i][j] + A[i][k] * B[k][j];
end;


{In this function, A is the coefficient matrix and B is the constant matrix (in your case, the transposed matrix ATb).
The function performs Gaussian elimination with partial pivoting to transform A into an upper triangular matrix,
and then it performs back substitution to find the solution vector.

{Gaussian elimination works well for many cases, but it's important to note that it may not be the best choice for all scenarios,
especially if the matrix A is ill-conditioned or nearly singular. }

function SolveLinearSystem2(A: TMatrix; B: TVector): TVector;
var
  n, i, j, k, maxRow: Integer;
  tmp, f: Double;
begin
  n := Length(A);
  SetLength(Result, n);

  // Forward elimination
  for i := 0 to n - 1 do begin
    // Find pivot row
    maxRow := i;
    for k := i + 1 to n - 1 do
      if Abs(A[k][i]) > Abs(A[maxRow][i]) then
        maxRow := k;

    // Swap rows
    for k := i to n - 1 do begin
      tmp := A[maxRow][k];
      A[maxRow][k] := A[i][k];
      A[i][k] := tmp;
    end;
    tmp := B[maxRow];
    B[maxRow] := B[i];
    B[i]:= tmp;



    // Make all rows below this one 0 in current column
    for k := i + 1 to n - 1 do begin
      f := A[k][i] / A[i][i];
      for j := i to n - 1 do
        A[k][j] := A[k][j] - A[i][j] * f;
      B[k] := B[k] - B[i] * f;
    end;
  end;

  // Back substitution
  for i := n - 1 downto 0 do begin
    Result[i] := B[i];
    for j := i + 1 to n - 1 do
      Result[i] := Result[i] - A[i][j] * Result[j];
    Result[i] := Result[i] / A[i][i];
  end;
end;


function SolveLinearSystem(A: TMatrix; B: TMatrix): TVector;
var
  n, i, j, k, maxRow: Integer;
  tmp, f: Double;
begin
  n := Length(A);
  SetLength(Result, n);

  // Forward elimination
  for i := 0 to n - 1 do begin
    // Find pivot row
    maxRow := i;
    for k := i + 1 to n - 1 do
      if Abs(A[k][i]) > Abs(A[maxRow][i]) then
        maxRow := k;

    // Swap rows
    for k := i to n - 1 do begin
      tmp := A[maxRow][k];
      A[maxRow][k] := A[i][k];
      A[i][k] := tmp;
    end;
    tmp := B[maxRow][0];
    B[maxRow][0] := B[i][0];
    B[i][0] := tmp;



    // Make all rows below this one 0 in current column
    for k := i + 1 to n - 1 do begin
      f := A[k][i] / A[i][i];
      for j := i to n - 1 do
        A[k][j] := A[k][j] - A[i][j] * f;
      B[k][0] := B[k][0] - B[i][0] * f;
    end;
  end;

  // Back substitution
  for i := n - 1 downto 0 do begin
    Result[i] := B[i][0];
    for j := i + 1 to n - 1 do
      Result[i] := Result[i] - A[i][j] * Result[j];
    Result[i] := Result[i] / A[i][i];
  end;
end;




{This function performs in-place modifications on the matrix A and constructs the inverse matrix in Result.
Note that this implementation does not handle cases where the matrix is singular (not invertible) or nearlysingular.
In practice, additional checks and more robust numerical methods are often used to handle these cases and to
improve the numerical stability of the inversion process.}
function InvertMatrix(A: TMatrix): TMatrix;
var
  n, i, j, k: Integer;
  factor, temp: Double;
begin
  n := Length(A);
  Result := CreateMatrix(n, n);

  // Initialize Result as the identity matrix
  for i := 0 to n - 1 do begin
    for j := 0 to n - 1 do
      if i = j then
        Result[i][j] := 1
      else
        Result[i][j] := 0;
  end;

  // Apply Gauss-Jordan elimination
  for i := 0 to n - 1 do begin
    // Make the diagonal element 1
    factor := A[i][i];
    for j := 0 to n - 1 do begin
      A[i][j] := A[i][j] / factor;
      Result[i][j] := Result[i][j] / factor;
    end;

    // Make the off-diagonal elements 0
    for j := 0 to n - 1 do begin
      if i <> j then begin
        factor := A[j][i];
        for k := 0 to n - 1 do begin
          A[j][k] := A[j][k] - A[i][k] * factor;
          Result[j][k] := Result[j][k] - Result[i][k] * factor;
        end;
      end;
    end;
  end;
end;


{This function takes two vectors (Vec1 and Vec2) as input and returns their dot product.
It first checks if the vectors are of equal length, as the dot product is only defined for vectors of the same dimension.
Then, it iterates through the vectors, multiplying the corresponding elements and accumulating the sum.}
function DotProduct(const Vec1, Vec2: TVector): Double;
var
  i: Integer;
begin
  if Length(Vec1) <> Length(Vec2) then
    raise Exception.Create('Vector lengths must be equal.');

  Result := 0.0;
  for i := 0 to High(Vec1) do
    Result := Result + Vec1[i] * Vec2[i];
end;


function FlattenMatrix(const M: TMatrix): TVector;
var
  i, j, k, Rows, Cols: Integer;
begin
  Rows := Length(M);
  Cols := Length(M[0]);
  SetLength(Result, Rows * Cols);
  k := 0;
  for i := 0 to Rows - 1 do
    for j := 0 to Cols - 1 do
    begin
      Result[k] := M[i][j];
      Inc(k);
    end;
end;

//The ConvertVectorToMatrix function is a helper function to convert the one-dimensional b vector into a two-dimensional matrix form,
//  which is needed for matrix multiplication.
function ConvertVectorToMatrix(const Vec: TVector): TMatrix;
var
  i: Integer;
begin
  SetLength(Result, Length(Vec), 1);
  for i := 0 to High(Vec) do
    Result[i][0] := Vec[i];
end;



//GeneratePolynomialTerms function dynamically generates the terms of the polynomial for a given order.
function GeneratePolynomialTerms(x, y: Double; order: Integer): TVector;
var
  i, j, k: Integer;
begin
  SetLength(Result, ((order + 1) * (order + 2)) div 2);
  k := 0;
  for i := 0 to order do
    for j := 0 to order - i do
    begin
      Result[k] := Power(x, i) * Power(y, j);
      Inc(k);
    end;
end;

{

LeastSquares function now takes an additional parameter order, which specifies the order of the polynomial.

This approach allows you to easily adjust the polynomial order by changing the order parameter when calling LeastSquares.
The design matrix A will be constructed accordingly with the appropriate number of terms for the specified polynomial order.

The function constructs the design matrix A using SampleData.
It transposes A to get AT.
It calculates ATA and ATb.
Finally, it calls SolveLinearSystem to solve the least squares problem.}

function LeastSquares(const SampleData: TMatrix; const b: TVector; order: Integer): TVector;
var
  A, AT, ATA, ATb: TMatrix;
  i, termCount,samples: Integer;
  x, y: Double;
begin
  termCount := ((order + 1) * (order + 2)) div 2; // Number of terms in the polynomial
  samples:=length(b);//A and b are the same size
  A := CreateMatrix(samples, termCount);

  for i := 0 to samples - 1 do
  begin
    x := SampleData[0][i]; // x-coordinate
    y := SampleData[1][i]; // y-coordinate
    A[i] := GeneratePolynomialTerms(x, y, order);
  end;

  // Transpose the matrix A
  AT := TransposeMatrix(A);

  // Calculate ATA (A transposed times A)
  ATA := MultiplyMatrices(AT, A);

  // Calculate ATb (A transposed times b)
  ATb := MultiplyMatrices(AT, ConvertVectorToMatrix(b));

  // Solve the least squares problem (find x in Ax = b)
  Result := SolveLinearSystem(ATA, ATb);
end;







// Program test ===========================================================================================


procedure GivensRotation(a, b: Double; out c, s: Double);
var
  h: Double;
begin
  if b <> 0.0 then
  begin
    h := Hypot(a, b); // Hypot computes the hypotenuse of a right-angled triangle
    c := a / h;
    s := -b / h;
  end
  else
  begin
    c := 1.0;
    s := 0.0;
  end;
end;


procedure GivensSolution(var A: TMatrix; var b: TVector; out x: TVector);
var
  n, i, j, k: Integer;
  c, s, temp: Double;
begin
  n := Length(A);
  SetLength(x, n);

  // Transform A into upper triangular form using Givens rotations
  for i := 0 to n - 2 do
    for j := i + 1 to n - 1 do
    begin
      if A[j][i] <> 0.0 then
      begin
        GivensRotation(A[i][i], A[j][i], c, s);

        // Apply the rotation to the matrix A and vector b
        for k := i to n - 1 do
        begin
          temp := c * A[i][k] - s * A[j][k];
          A[j][k] := s * A[i][k] + c * A[j][k];
          A[i][k] := temp;
        end;

        temp := c * b[i] - s * b[j];
        b[j] := s * b[i] + c * b[j];
        b[i] := temp;
      end;
    end;

  // Back substitution
  for i := n - 1 downto 0 do
  begin
    x[i] := b[i];
    for j := i + 1 to n - 1 do
      x[i] := x[i] - A[i][j] * x[j];
    if A[i][i] = 0.0 then
      raise Exception.Create('Matrix is singular.');
    x[i] := x[i] / A[i][i];
  end;
end;


function GaussianEliminationAswapped(A: TMatrix; B: TVector): TVector;
var
  n, i, j, k, maxRow: Integer;
  factor, temp, maxPivot: Double;
begin
  n := Length(A); // Assuming all columns have the same number of rows
  SetLength(Result, n);

  // Forward elimination with partial pivoting
  for i := 0 to n - 2 do
  begin

    // Find the pivot row
    maxPivot := Abs(A[i][i]);
    maxRow := i;
    for k := i + 1 to n - 1 do
      if Abs(A[i][k]) > maxPivot then
      begin
        maxPivot := Abs(A[i][k]);
        maxRow := k;
      end;

    // Check if the pivot is too small (near zero)
    if maxPivot < 1E-10 then
      raise Exception.Create('Matrix is singular or nearly singular.');


    // Swap the rows if necessary
    for k := i to n - 1 do
    begin
      temp := A[k][maxRow];
      A[k][maxRow] := A[k][i];
      A[k][i] := temp;
    end;
    temp := B[maxRow];
    B[maxRow] := B[i];
    B[i] := temp;


    // Perform elimination
    for j := i + 1 to n - 1 do
    begin
      factor := A[i][j] / A[i][i];
      for k := i to n - 1 do
        A[k][j] := A[k][j] - factor * A[k][i];
      B[j] := B[j] - factor * B[i];
    end;
  end;


  // Back substitution
  for i := n - 1 downto 0 do
  begin
    Result[i] := B[i];
    for j := i + 1 to n - 1 do
      Result[i] := Result[i] - A[j][i] * Result[j];
    if A[i][i] = 0 then
      raise Exception.Create('Matrix is singular or nearly singular.');
    Result[i] := Result[i] / A[i][i];
  end;

end;


function GaussianElimination(A: TMatrix; B: TVector): TVector;
var
  n, i, j, k, maxRow: Integer;
  factor, temp, maxPivot: Double;
begin
  n := Length(A[0]);
  SetLength(Result, n);

  // Forward elimination with partial pivoting
  for i := 0 to n - 2 do
  begin

    // Find the pivot row
    maxPivot := Abs(A[i][i]);
    maxRow := i;
    for k := i + 1 to n - 1 do
      if Abs(A[k][i]) > maxPivot then
      begin
        maxPivot := Abs(A[k][i]);
        maxRow := k;
      end;

    // Check if the pivot is too small (near zero)
    if maxPivot < 1E-10 then
      raise Exception.Create('Matrix is singular or nearly singular.');

    // Swap the rows if necessary
    for k := i to n - 1 do
    begin
      temp := A[maxRow][k];
      A[maxRow][k] := A[i][k];
      A[i][k] := temp;
    end;
    temp := B[maxRow];
    B[maxRow] := B[i];
    B[i] := temp;

    // Perform elimination
    for j := i + 1 to n - 1 do
    begin
      factor := A[j][i] / A[i][i];
      for k := i to n - 1 do
        A[j][k] := A[j][k] - factor * A[i][k];
      B[j] := B[j] - factor * B[i];

    end;
  end;

  // Back substitution
  for i := n - 1 downto 0 do
  begin
    Result[i] := B[i];
    for j := i + 1 to n - 1 do
      Result[i] := Result[i] - A[i][j] * Result[j];
    if A[i][i] = 0 then
      raise Exception.Create('Matrix is singular or nearly singular.');
    Result[i] := Result[i] / A[i][i];
  end;
end;


{   lsq_fit:                                                                                                                                     }
{   Find the solution vector of an overdetermined system of linear equations according to the method of least squares using GIVENS rotations     }
{                                                                                                                                                }
{   Solve x of A x = b with the least-squares method                                                                                             }
{   In matrix calculations, b_matrix[0..nr_columns-1,0..nr_equations-1]:=solution_vector[0..2] * A_XYpositions[0..nr_columns-1,0..nr_equations-1]}
{                                                                                                                              }
{   see also Montenbruck & Pfleger, Astronomy on the personal computer}
procedure lsq_fit(const A_matrix: TMatrix; {[, 0..3,0..nr_equations-1]} b_matrix  : TVector;{equations result, b=A*s}  out x_matrix: TVector );
  const tiny = 1E-10;  {accuracy}
  var i,j,k, nr_equations,nr_columns,hhh,n  : integer;
      p,q,h                           : double;
      temp_matrix                     : TMatrix;

begin
  n := Length(A_matrix);
  SetLength(x_matrix, n);


  nr_equations:=length(A_matrix[0]);
  nr_columns:=length(A_matrix);{should be 3 for this application}

  temp_matrix:=A_matrix; {In dynamic arrays, the assignment statement duplicates only the reference to the array, while SetLength does the job of physically copying/duplicating it, leaving two separate, independent dynamic arrays.}
  setlength(temp_matrix,nr_columns,nr_equations);{duplicate A_matrix to protect data in A_matrix}

  for j:=0 to nr_columns-1 do {loop over columns of temp_matrix}
  {eliminate matrix elements A[i,j] with i>j from column j}
    for i:=j+1 to nr_equations-1 do
      if temp_matrix[j,i]<>0 then
      begin{calculate p, q and new temp_matrix[j,j]; set temp_matrix[j,i]=0}
        if abs(temp_matrix[j,j])<tiny*abs(temp_matrix[j,i]) then
        begin
          p:=0;
          q:=1;
          temp_matrix[j,j]:=-temp_matrix[j,i];
          temp_matrix[j,i]:=0;

        end
        else
        begin
          h:=sqrt(temp_matrix[j,j]*temp_matrix[j,j]+temp_matrix[j,i]*temp_matrix[j,i]);
          if temp_matrix[j,j]<0 then h:=-h;
          p:=temp_matrix[j,j]/h;
          q:=-temp_matrix[j,i]/h;
          temp_matrix[j,j]:=h;
          temp_matrix[j,i]:=0;

        end;
        {calculate the rest of the line}
        for k:=j+1 to nr_columns-1 do
        begin
          h:= p*temp_matrix[k,j] - q*temp_matrix[k,i];
          temp_matrix[k,i] := q*temp_matrix[k,j] + p*temp_matrix[k,i];
          temp_matrix[k,j] := h;
        end;
        h:= p*b_matrix[j] - q*b_matrix[i];
        b_matrix[i] := q*b_matrix[j] + p*b_matrix[i];
        b_matrix[j] := h;
      end;

  for i:=0 to nr_columns-1 do x_matrix[i]:=0; //2022, extra precaution to zero x_matrix

  for i:= nr_columns-1 downto 0 do {back substitution}
  begin
    H:=b_matrix[i];
    for k:=i+1 to nr_columns-1 do
            h:=h-temp_matrix[k,i]*x_matrix[k];
    if temp_matrix[i,i]<>0 then x_matrix[i] := h/temp_matrix[i,i]
    else
    x_matrix[i]:=9999999999999999; //v2022. Prevent runtime error dividing by zero. Should normally not happen. In case of zero due to wrong double star detection by using triples force a failure

    {solution vector x:=x_matrix[0]x+x_matrix[1]y+x_matrix[2]}
  end;
end; {lsq_fit}


procedure rotate(rot,x,y :double;var  x2,y2:double);{rotate a vector point, angle seen from y-axis, counter clockwise}
var
  sin_rot, cos_rot :double;
begin
  sincos(rot, sin_rot, cos_rot);
  x2:=x * + sin_rot + y*cos_rot;{ROTATION MOON AROUND CENTER OF PLANET}
  y2:=x * - cos_rot + y*sin_rot;{SEE PRISMA WIS VADEMECUM BLZ 68}
end;

var
  SampleData,atb,asw :TMatrix;
  b,b2,c, coefficients,coefficients2,coefficients3,xx: TVector;
  i,j,k, nrstars,count: Integer;
  centerX, centerY, dx, dy, r, RMSE, err, offset, dr,x,y,test,x2,y2: Double;

  a :tmatrix;
  stars : array[0..1,0..1000] of double;
  errors : array[0..1000] of double;
  f,g,h :string;


begin
  count:=0;
  setlength(asw,4,1000);
  setlength(b,1000);
  setlength(c,1000);
  for i:=-10 to +10 do
  for j:=-10 to 10 do
  if ((j<>0) and (i<>0)) then
  begin

    y:=i;
    x:=j;

    b[count]:=j+1000;//reference X
    c[count]:=j+1000;

    rotate(pi*30/180,x,y, x2,y2);
    x:=x2;
    y:=y2;

    x:=x+0.01*x*x;//process
     //+0.01*y*y;//process
    asw[0,count]:=x;
    asw[1,count]:=y;
    asw[2,count]:=1;
    asw[3,count]:=x*x;
{    asw[4,count]:=0;
    asw[5,count]:=0;
    asw[6,count]:=0;
    asw[7,count]:=0;
    asw[8,count]:=0;
    asw[9,count]:=0;}

    inc(count);
  end;
  setlength(asw,4,count);
  setlength(b,count);
  lsq_fit(Asw, b, coefficients3);
  count:=2;

  x:=asw[0,count];
  y:=asw[1,count];

  test:=coefficients3[0]*x  +coefficients3[1]*y + coefficients3[2] + x*x*coefficients3[3]{ - x*x*coefficients3[4]  - y*y*coefficients3[5]} -   c[count];

  exit;

  setlength(a,6,3);
  a[0,0]:=208;
  a[0,1]:=595;
  a[0,2]:=1;


  a[1,0]:=210;
  a[1,1]:=674;
  a[1,2]:=1;


  a[2,0]:=204;
  a[2,1]:=645;
  a[2,2]:=1;

  a[3,0]:=208;
  a[3,1]:=712;
  a[3,2]:=1;

  a[4,0]:=253;
  a[4,1]:=483;
  a[4,2]:=1;

  a[5,0]:=318;
  a[5,1]:=68;
  a[5,2]:=1;

  setlength(asw,3,6);
  asw[0,0]:=208;
  asw[1,0]:=595;
  asw[2,0]:=1;


  asw[0,1]:=210;
  asw[1,1]:=679;
  asw[2,1]:=1;


  asw[0,2]:=204;
  asw[1,2]:=645;
  asw[2,2]:=1;

  asw[0,3]:=208;
  asw[1,3]:=712;
  asw[2,3]:=1;

  asw[0,4]:=253;
  asw[1,4]:=483;
  asw[2,4]:=1;

  asw[0,5]:=318;
  asw[1,5]:=68;
  asw[2,5]:=1;

  setlength(b,6);
  b[0]:=-1625;
  b[1]:=-1597;
  b[2]:=-1621;
  b[3]:=-1596;
  b[4]:=-1538;
  b[5]:=-1570;



{
  a[0,0]:=204;
  a[1,0]:=645;
  a[2,0]:=1;

  a[0,1]:=253;
  a[1,1]:=483;
  a[2,1]:=1;

  a[0,2]:=318;
  a[1,2]:=68;
  a[2,2]:=1;}

{  a[0,0]:=204;
  a[0,1]:=645;
  a[0,2]:=1;

  a[1,0]:=253;
  a[1,1]:=483;
  a[1,2]:=1;

  a[2,0]:=318;
  a[2,1]:=68;
  a[2,2]:=1;   }

{  setlength(b,3);
  b[0]:=-1621;
  b[1]:=-1538;
  b[2]:=-1470; }

//  lsq_fit(Asw, b, coefficients3);

// GivensSolution(A, b, coefficients2);

//  coefficients3:=SolveLinearSystem2(A,B);



  setlength(asw,3,500);
  setlength(a,500,3);
  for i:=6 to 500-1 do
  begin
    asw[0,i]:=asw[0,i-6];
    asw[1,i]:=asw[1,i-6];
    asw[2,i]:=asw[2,i-6];
    a[i,0]:=a[i-6,0];
    a[i,1]:=a[i-6,1];
    a[i,2]:=a[i-6,2];

  end;

//  setlength(b2,lenght(b);

  f:=timetostr(time);


//  for i:=1 to 400000000 do
//  coefficients:=GaussianElimination(A,B);
//  GivensSolution(A, b, coefficients2);

  g:=timetostr(time);


  for i:=1 to 400000000 do
  begin
//  coefficients2:=GaussianEliminationAswapped(Asw,B);

    b2:=copy(b,0);
    lsq_fit(Asw, b2, coefficients3);
  end;
  h:=timetostr(time);


   exit;


  k:=0;
  for i:=1 to 7 do
  for j:=1 to 7 do
  begin
    if ((j<>0) and (i<>0)) then
    begin
      x:=j;
      y:=i;
      stars[0,k]:=x;
      stars[1,k]:=y;
      Errors[k]:=0.001+0.001*abs(x){+0.001*abs(y)+0.002*abs(x*y)+0.003*x*x+0.003*y*y +0.0005*abs(x*x*y)+0.00005*abs(x*x*y)};
      inc(k);
    end;
  end;
  nrstars:=k;

//  Each row in the matrix A corresponds to one of the data points (the grid points),
//  and each column corresponds to one of the polynomial terms.
//  The matrix A will thus have 9 rows (one for each point in the 3x3 grid) and 10 columns
// (one for each coefficient in the polynomial):

  Setlength(A, nrstars, 3{10}); // 9 points and 10 terms in the polynomial

//  Vector BB contains the distortion values (the zz-values) for each point in the 3x3 grid:
  Setlength(B, nrstars); // 9 points

//  Let's fill in the matrix AA with the values based on the 3x3 grid centered at the origin,
//  assuming x and y coordinates range from -1 to 1. We will also fill in the vector BB with
//  the error values you've provided:

  // Assuming the center of the grid is at (0,0)
  for i := 0 to nrstars-1 {8} do
  begin
    x := stars[0,i];
    y := stars[1,i];

    // Fill the design matrix A
    // Assuming i is correctly iterating over the dataset and x and y are correctly set
    A[i][0] := 1;         // This is for the constant term a0
    A[i][1] := x;         // This is for the term a1*x
    A[i][2] := y;         // This is for the term a2*y
{    A[i][3] := x * y;     // This is for the term a3*x*y
    A[i][4] := x * x;     // This is for the term a4*x^2
    A[i][5] := y * y;     // This is for the term a5*y^2
    A[i][6] := x * x * x; // This is for the term a6*x^3
    A[i][7] := x * x * y; // This is for the term a7*x^2*y
    A[i][8] := y * y * y; // This is for the term a8*y^3
    A[i][9] := x * y * y; // This is for the term a9*x*y^2
 }
    // Fill in the observed errors into vector B
     B[i] := errors[i];
  end;

  coefficients:=SolveLinearSystem2(A, B );
//  coefficients:=GaussianElimination2(A,B);
//  GivensSolution(A, b, coefficients);
  coefficients:=GaussianElimination(A,B);

  beep;
  // coefficients now contains the estimated polynomial coefficients
  //  a0​ (constant term)
  //  + a1*x, a2*y
  //  + a3*x*y, a4*x^2, a5*y^2
  //  + a6*x^3, a7*x^2*y, a8*x*y^2, a9y^3   ==> 3th Order Polynomial
  //  + a10*x^4, a11*x^3*y, a12*x^2*y^2, a13*x*y^3, a14*y^4     ==> 4th Order Polynomial
  //  + a15*x^5, a16*x^4*y, a17*x^3*y^2, a18*x^2*y^3, a198x8y^4, a20*y^5  ==> 5th Order Polynomial
  beep;
end.



//https://web.mit.edu/10.001/Web/Course_Notes/GaussElimPivoting.html
{  setlength(a,4,4);
a[0,0]:=0.02;
a[0,1]:=0.01;
a[0,2]:=0;
a[0,3]:=0;

a[1,0]:=1;
a[1,1]:=2;
a[1,2]:=1;
a[1,3]:=0;

a[2,0]:=0;
a[2,1]:=1;
a[2,2]:=2;
a[2,3]:=1;

a[3,0]:=0;
a[3,1]:=0;
a[3,2]:=100;
a[3,3]:=200;


setlength(atb,4,1);
atb[0,0]:=0.02;
atb[1,0]:=1;
atb[2,0]:=4;
atb[3,0]:=800;
xx := SolveLinearSystem(A, ATb);

//  Step 8: Back Substitute
// -0.2x4 = -0.05;    x4 = 4
// 100x3 + 200x4 = 800;   x3 = 0
// x2 + 2x3 + x4 =4;   x2 = 0
// x1 + 2x2 + x3 = 1;   x1 = 1
}


{  Setlength(A, 7,7);
  Setlength(B, 7);

  a[0,0]:=360360;
  a[0,1]:=180180;
  a[0,2]:=120120;
  a[0,3]:=90090;
  a[0,4]:=72072;
  a[0,5]:=60060;
  a[0,6]:=51480;

  a[1,0]:=180180;
  a[1,1]:=120120;
  a[1,2]:=90090;
  a[1,3]:=72072;
  a[1,4]:=60060;
  a[1,5]:=51480;
  a[1,6]:=45045;

  a[2,0]:=120120;
  a[2,1]:=90090;
  a[2,2]:=72072;
  a[2,3]:=60060;
  a[2,4]:=51480;
  a[2,5]:=45045;
  a[2,6]:=40040;

  a[3,0]:=90090;
  a[3,1]:=72072;
  a[3,2]:=60060;
  a[3,3]:=51480;
  a[3,4]:=45045;
  a[3,5]:=40040;
  a[3,6]:=36036;

  a[4,0]:=72072;
  a[4,1]:=60060;
  a[4,2]:=51480;
  a[4,3]:=45045;
  a[4,4]:=40040;
  a[4,5]:=36036;
  a[4,6]:=32760;

  a[5,0]:=60060;
  a[5,1]:=51480;
  a[5,2]:=45045;
  a[5,3]:=40040;
  a[5,4]:=36036;
  a[5,5]:=32760;
  a[5,6]:=30030;

  a[6,0]:=51480;
  a[6,1]:=45045;
  a[6,2]:=40040;
  a[6,3]:=36036;
  a[6,4]:=32760;
  a[6,5]:=30030;
  a[6,6]:=27720;

  b[0]:=273702;
  b[1]:=131703;
  b[2]:=88517;
  b[3]:=67639;
  b[4]:=55211;
  b[5]:=46891;
  b[6]:=40889;

  GivensSolution(A, b, coefficients2);
  coefficients:=GaussianElimination(A,B);
  //Coefficients are 1,-1 ,1,-1,1,-1,1
}


